# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import savanna.openstack.common.importutils as importutils

_CONF = importutils.try_import('savanna.tests.integration.configs.hdp_config')


def _get_conf(key, default):
    return getattr(_CONF, key) if _CONF and hasattr(_CONF, key) else default

PLUGIN_NAME = _get_conf('PLUGIN_NAME', 'hdp')

IMAGE_ID = _get_conf('IMAGE_ID', '5ea141c3-893e-4b5c-b138-910adc09b281')

NODE_USERNAME = _get_conf('NODE_USERNAME', 'cloud-user')

HADOOP_VERSION = _get_conf('HADOOP_VERSION', '1.3.0')
HADOOP_USER = _get_conf('HADOOP_USER', 'hdfs')
HADOOP_DIRECTORY = _get_conf('HADOOP_DIRECTORY', '/usr/lib/hadoop')
HADOOP_LOG_DIRECTORY = _get_conf('HADOOP_LOG_DIRECTORY',
                                 '/hadoop/mapred/userlogs')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import savanna.openstack.common.importutils as importutils

_CONF = importutils.try_import(
    'savanna.tests.integration.configs.common_config')


def _get_conf(key, default):
    return getattr(_CONF, key) if _CONF and hasattr(_CONF, key) else default

OS_USERNAME = _get_conf('OS_USERNAME', 'admin')
OS_PASSWORD = _get_conf('OS_PASSWORD', 'admin')
OS_TENANT_NAME = _get_conf('OS_TENANT_NAME', 'admin')
OS_AUTH_URL = _get_conf('OS_AUTH_URL', 'http://127.0.0.1:35357/v2.0/')

SAVANNA_HOST = _get_conf('SAVANNA_HOST', '127.0.0.1')
SAVANNA_PORT = _get_conf('SAVANNA_PORT', '8386')

FLAVOR_ID = _get_conf('FLAVOR_ID', '2')

TIMEOUT = _get_conf('TIMEOUT', 45)

CLUSTER_NAME = _get_conf('CLUSTER_NAME', 'test-cluster')

USER_KEYPAIR_ID = _get_conf('USER_KEYPAIR_ID', 'jenkins')

PATH_TO_SSH = _get_conf('PATH_TO_SSH', '/home/ubuntu/.ssh/id_rsa')

JT_PORT = _get_conf('JT_PORT', 50030)
NN_PORT = _get_conf('NN_PORT', 50070)
TT_PORT = _get_conf('TT_PORT', 50060)
DN_PORT = _get_conf('DN_PORT', 50075)
SEC_NN_PORT = _get_conf('SEC_NN_PORT', 50090)

ENABLE_CLUSTER_CL_TEMPLATE_CRUD_TESTS = _get_conf(
    'ENABLE_CLUSTER_CL_TEMPLATE_CRUD_TESTS', True)
ENABLE_CLUSTER_NGT_NODE_PROCESS_CRUD_TESTS = _get_conf(
    'ENABLE_CLUSTER_NGT_NODE_PROCESS_CRUD_TESTS', True)
ENABLE_CLUSTER_NGT_CRUD_TESTS = _get_conf(
    'ENABLE_CLUSTER_NGT_CRUD_TESTS', True)
ENABLE_CLUSTER_NODE_PROCESS_CRUD_TESTS = _get_conf(
    'ENABLE_CLUSTER_NODE_PROCESS_CRUD_TESTS', True)

ENABLE_CL_TEMPLATE_CRUD_TESTS = _get_conf(
    'ENABLE_CL_TEMPLATE_CRUD_TESTS', True)
ENABLE_NGT_CRUD_TESTS = _get_conf('ENABLE_NGT_CRUD_TESTS', True)

ENABLE_HADOOP_TESTS_FOR_VANILLA_PLUGIN = _get_conf(
    'ENABLE_HADOOP_TESTS_FOR_VANILLA_PLUGIN', True)
ENABLE_HADOOP_TESTS_FOR_HDP_PLUGIN = _get_conf(
    'ENABLE_HADOOP_TESTS_FOR_HDP_PLUGIN', True)
ENABLE_SWIFT_TESTS = _get_conf('ENABLE_SWIFT_TESTS', True)
ENABLE_SCALING_TESTS = _get_conf('ENABLE_SCALING_TESTS', True)
ENABLE_CONFIG_TESTS = _get_conf('ENABLE_CONFIG_TESTS', True)

ENABLE_IR_TESTS = _get_conf('ENABLE_IR_TESTS', True)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base
import savanna.tests.integration.configs.parameters.common_parameters as param
import savanna.tests.integration.configs.parameters.vanilla_parameters as v_prm


NAMENODE_CONFIG = {'Name Node Heap Size': 510}
JOBTRACKER_CONFIG = {'Job Tracker Heap Size': 512}

DATANODE_CONFIG = {'Data Node Heap Size': 511}
TASKTRACKER_CONFIG = {'Task Tracker Heap Size': 513}

GENERAL_CONFIG = {'Enable Swift': True}

CLUSTER_HDFS_CONFIG = {'dfs.replication': 2}
CLUSTER_MAPREDUCE_CONFIG = {'mapred.map.tasks.speculative.execution': False,
                            'mapred.child.java.opts': '-Xmx500m'}


def _add_config(body, config_type, service, config):
    body[config_type][service] = config


@base.enable_test(param.ENABLE_CONFIG_TESTS)
class ClusterConfigTest(base.ITestCase):

    def setUp(self):
        super(ClusterConfigTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

    def assertConfigs(self, get_config, param_config):
        self.assertEqual(get_config, param_config,
                         'Configs are not equal: \'%s\' != \'%s\''
                         % (str(get_config), str(param_config)))

    def assertConfigOnNode(self, host, config, value):
        conf = config.replace(' ', '')
        com = self.execute_command(host, './script.sh %s -val %s -url %s'
                                         % (conf, value, param.OS_AUTH_URL),
                                   v_prm.NODE_USERNAME)
        self.assertEqual(com[0], 0,
                         'Host: %s, config \'%s\' is not equal \'%s\''
                         % (host, config, value))

    def _cluster_config_testing(self, cluster_id):
        get_data = self.get_object(self.url_cluster_with_slash,
                                   cluster_id, 200, True)
        get_data = get_data['cluster']

        self.assertConfigs(get_data['cluster_configs']['general'],
                           GENERAL_CONFIG)
        self.assertConfigs(get_data['cluster_configs']['HDFS'],
                           CLUSTER_HDFS_CONFIG)
        self.assertConfigs(get_data['cluster_configs']['MapReduce'],
                           CLUSTER_MAPREDUCE_CONFIG)

        process_map = {
            'namenode': {
                'service': 'HDFS', 'param': NAMENODE_CONFIG
            },
            'jobtracker': {
                'service': 'MapReduce',
                'param': JOBTRACKER_CONFIG
            },
            'datanode': {
                'service': 'HDFS', 'param': DATANODE_CONFIG
            },
            'tasktracker': {
                'service': 'MapReduce', 'param': TASKTRACKER_CONFIG
            }
        }

        def get_node_configs(node_group, process):
            return \
                node_group['node_configs'][process_map[process]['service']]

        def get_param(process):
            return process_map[process]['param']

        node_groups = get_data['node_groups']
        ip_instances = {}
        for node_group in node_groups:
            for process in node_group['node_processes']:
                self.assertConfigs(
                    get_node_configs(node_group, process),
                    get_param(process)
                )

            instances = node_group['instances']
            for instans in instances:
                management_ip = instans['management_ip']
                self.transfer_script_to_node(
                    management_ip, v_prm.NODE_USERNAME,
                    'hadoop_config_test/hadoop_config_test_script.sh'
                )
                ip_instances[management_ip] = node_group[
                    'node_processes']

        for key, processes in ip_instances.items():
            telnetlib.Telnet(key, '22')

            for conf, value in CLUSTER_MAPREDUCE_CONFIG.items():
                self.assertConfigOnNode(key, conf, value)

            for conf, value in CLUSTER_HDFS_CONFIG.items():
                self.assertConfigOnNode(key, conf, value)

            for process in processes:
                for sec_key, sec_value in get_param(process).items():
                    self.assertConfigOnNode(key, sec_key, sec_value)

            if 'namenode' in processes:
                for sec_key, sec_value in GENERAL_CONFIG.items():
                    self.assertConfigOnNode(key, sec_key, sec_value)

    def test_cluster_config_nnjt_ttdn(self):
        master_ngt_body = self.make_node_group_template(
            'master-ngt', 'qa probe', 'JT+NN')
        _add_config(master_ngt_body, 'node_configs', 'HDFS', NAMENODE_CONFIG)
        _add_config(master_ngt_body, 'node_configs', 'MapReduce',
                    JOBTRACKER_CONFIG)

        worker_ngt_body = self.make_node_group_template(
            'worker-ngt', 'qa probe', 'TT+DN')
        _add_config(worker_ngt_body, 'node_configs', 'HDFS', DATANODE_CONFIG)
        _add_config(worker_ngt_body, 'node_configs', 'MapReduce',
                    TASKTRACKER_CONFIG)

        try:
            id_master_ngt = self.get_object_id(
                'node_group_template', self.post_object(self.url_ngt,
                                                        master_ngt_body, 202)
            )
        except Exception as e:
            self.fail('Failure while master node group template creation: ' +
                      str(e))

        try:
            id_worker_ngt = self.get_object_id(
                'node_group_template', self.post_object(self.url_ngt,
                                                        worker_ngt_body, 202)
            )
        except Exception as e:
            self.del_object(self.url_ngt_with_slash, id_master_ngt, 204)
            self.fail('Failure while worker node group template creation: ' +
                      str(e))

        ngt_id_list = {id_master_ngt: 1, id_worker_ngt: 2}
        cluster_body = self.make_cl_body_node_group_templates(ngt_id_list)

        _add_config(cluster_body, 'cluster_configs', 'general', GENERAL_CONFIG)
        _add_config(cluster_body, 'cluster_configs', 'HDFS',
                    CLUSTER_HDFS_CONFIG)
        _add_config(cluster_body, 'cluster_configs', 'MapReduce',
                    CLUSTER_MAPREDUCE_CONFIG)

        try:
            cluster_id = self.create_cluster_and_get_id(cluster_body)

        except Exception as e:
            self.del_object(self.url_ngt_with_slash, id_master_ngt, 204)
            self.del_object(self.url_ngt_with_slash, id_worker_ngt, 204)
            self.fail('Failure while cluster creation: ' + str(e))

        try:
            self._cluster_config_testing(cluster_id)

        except Exception as e:
            self.fail('Failure while config comparison: ' + str(e))

        finally:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)
            self.del_object(self.url_ngt_with_slash, id_master_ngt, 204)
            self.del_object(self.url_ngt_with_slash, id_worker_ngt, 204)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from novaclient import base
from novaclient.v1_1 import keypairs


class SavannaKeypairManager(keypairs.KeypairManager):
    def get(self, keypair):
        """Get a keypair.

        :param image: The ID of the keypair to get.
        :rtype: :class:`Keypair`
        """
        return self._get("/os-keypairs/%s" % base.getid(keypair), "keypair")



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base
import savanna.tests.integration.configs.parameters.common_parameters as param


@base.enable_test(param.ENABLE_CLUSTER_CL_TEMPLATE_CRUD_TESTS)
class ClusterFromClusterTemplateCrudTest(base.ITestCase):

    def setUp(self):
        super(ClusterFromClusterTemplateCrudTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

        self.create_node_group_templates()

    def crud_cluster_cl_template(self, node_list):
        cl_template_body = self.make_cluster_template('cl-template', node_list)

        try:
            cl_template_id = self.get_object_id(
                'cluster_template', self.post_object(self.url_cl_tmpl,
                                                     cl_template_body, 202))
        except Exception as e:
            self.fail('Failure while cluster template creation: ' + str(e))

        cluster_body = self.make_cl_body_cluster_template(cl_template_id)

        try:
            self.crud_object(cluster_body, self.url_cluster)

        except Exception as e:
            self.fail('CRUD test has failed: ' + str(e))

        finally:
            self.del_object(self.url_cl_tmpl_with_slash, cl_template_id, 204)

    def test_cluster_nnttdn_jt(self):
        """This test check cluster creation with topology | NN + TT + DN | JT |
        via cluster template.
        """
        node_list = {self.id_nn_tt_dn: 1, self.id_jt: 1}
        self.crud_cluster_cl_template(node_list)

    def test_cluster_jtttdn_nn(self):
        """This test check cluster creation with topology | JT + TT + DN | NN |
        via cluster template.
        """
        node_list = {self.id_jt_tt_dn: 1, self.id_nn: 1}
        self.crud_cluster_cl_template(node_list)

    def tearDown(self):
        self.delete_node_group_templates()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base
import savanna.tests.integration.configs.parameters.common_parameters as param


@base.enable_test(param.ENABLE_CLUSTER_NGT_CRUD_TESTS)
class ClusterNodeGroupTemplatesCrudTest(base.ITestCase):

    def setUp(self):
        super(ClusterNodeGroupTemplatesCrudTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

        self.create_node_group_templates()

    def test_crud_cluster_ngt_jtnn_ttdn(self):
        """This test checks cluster creation with
        topology | JT + NN | TT + DN | via node group templates.
        """
        ngt_id_list = {self.id_jt_nn: 1, self.id_tt_dn: 2}
        body = self.make_cl_body_node_group_templates(ngt_id_list)
        self.crud_object(body, self.url_cluster)

    def test_crud_cluster_ngt_jt_nn_tt_dn(self):
        """This test checks cluster creation with
        topology | JT | NN | TT | DN | via node group templates.
        """
        ngt_id_list = {self.id_jt: 1, self.id_nn: 1, self.id_tt: 1,
                       self.id_dn: 1}
        body = self.make_cl_body_node_group_templates(ngt_id_list)
        self.crud_object(body, self.url_cluster)

    def tearDown(self):
        self.delete_node_group_templates()



# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import savanna.db.models as m
from savanna.openstack.common import log as logging

LOG = logging.getLogger(__name__)


def convert(cluster_template, normalized_config, config_resource):
    cluster_template.hadoop_version = normalized_config.hadoop_version
    for ng in normalized_config.node_groups:
        template_relation = m.TemplatesRelation(ng.name, ng.flavor,
                                                ng.node_processes,
                                                ng.count)
        cluster_template.node_groups.append(template_relation)
    for entry in normalized_config.cluster_configs:
        user_input = next((ui for ui in config_resource
                           if entry.config.name == ui.name), None)
        if user_input is not None:
            ci = entry.config
            # get the associated service dictionary
            service_dict = cluster_template.cluster_configs.get(
                entry.config.applicable_target, {})
            service_dict[ci.name] = entry.value
            cluster_template.cluster_configs[
                entry.config.applicable_target] = \
                service_dict
        else:
            LOG.debug('Template based input {0} is being filtered out as it '
                      'is not considered a user input'
                      .format(entry.config.name))

    LOG.debug('Cluster configs: {0}'.format(cluster_template.cluster_configs))

    return cluster_template


def get_host_role(host):
    if hasattr(host, 'role'):
        return host.role
    else:
        return host.node_group.name


def get_node_processes(host):
    if hasattr(host, 'node_processes'):
        return host.node_processes
    else:
        return host.node_group.node_processes



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import json
import os
import telnetlib
import time

import keystoneclient.v2_0
import requests
import unittest2

import savanna.tests.integration.configs.parameters.common_parameters as param
import savanna.tests.integration.configs.parameters.hdp_parameters as hdp_param
import savanna.tests.integration.configs.parameters.vanilla_parameters as v_prm
from savanna.utils import remote


def enable_test(test):
    def function(fun):
        if test:
            return fun
        else:
            return unittest2.skip('Test is disabled')(fun)

    return function


class ITestCase(unittest2.TestCase):

    def setUp(self):
        self.port = param.SAVANNA_PORT
        self.host = param.SAVANNA_HOST

        self.baseurl = 'http://' + self.host + ':' + self.port

        self.keystone = keystoneclient.v2_0.Client(
            username=param.OS_USERNAME,
            password=param.OS_PASSWORD,
            tenant_name=param.OS_TENANT_NAME,
            auth_url=param.OS_AUTH_URL
        )

        self.tenant = self.keystone.tenant_id
        self.token = self.keystone.auth_token

        self.url_version = '/'

        self.url_ngt = '/v1.0/%s/node-group-templates' % self.tenant
        self.url_ngt_with_slash = '/v1.0/%s/node-group-templates/'\
                                  % self.tenant

        self.url_cluster = '/v1.0/%s/clusters' % self.tenant
        self.url_cluster_with_slash = '/v1.0/%s/clusters/' % self.tenant

        self.url_cl_tmpl = '/v1.0/%s/cluster-templates' % self.tenant
        self.url_cl_tmpl_with_slash = '/v1.0/%s/cluster-templates/'\
                                      % self.tenant

        self.url_plugins = '/v1.0/%s/plugins' % self.tenant
        self.url_plugins_with_slash = '/v1.0/%s/plugins/' % self.tenant

        self.url_images = '/v1.0/%s/images' % self.tenant
        self.url_images_with_slash = '/v1.0/%s/images/' % self.tenant

#----------------------methods_for_CRUD_operations-----------------------------

    def _post(self, url, body):
        URL = self.baseurl + url
        resp = requests.post(URL, data=body, headers={
            'x-auth-token': self.token, 'Content-Type': 'application/json'})
        data = json.loads(resp.content) if resp.status_code == 202 \
            else resp.content
        print('URL = %s \n data = %s \n response = %s \n data = %s \n'
              % (URL, body, resp.status_code, data))
        return resp

    def _put(self, url, body):
        URL = self.baseurl + url
        resp = requests.put(URL, data=body, headers={
            'x-auth-token': self.token, 'Content-Type': 'application/json'})
        data = json.loads(resp.content)
        print('URL = %s \n data = %s \n response = %s \n data = %s \n'
              % (URL, body, resp.status_code, data))
        return resp

    def _get(self, url, printing):
        URL = self.baseurl + url
        resp = requests.get(URL, headers={'x-auth-token': self.token})
        if printing:
            print('URL = %s \n response = %s \n' % (URL, resp.status_code))
        if resp.status_code != 200:
            data = json.loads(resp.content)
            print('data= %s \n' % data)
        return resp

    def _delete(self, url):
        URL = self.baseurl + url
        resp = requests.delete(URL, headers={'x-auth-token': self.token})
        print('URL = %s \n response = %s \n' % (URL, resp.status_code))
        if resp.status_code != 204:
            data = json.loads(resp.content)
            print('data= %s \n' % data)
        return resp

    def post_object(self, url, body, code):
        post = self._post(url, json.dumps(body))
        self.assertEqual(post.status_code, code)
        data = json.loads(post.content)
        return data

    def put_object(self, url, object_id, body, code):
        data = self._put(url + object_id, json.dumps(body))
        self.assertEqual(data.status_code, code)
        data = json.loads(data.content)
        return data

    def get_object(self, url, obj_id, code, printing=False):
        rv = self._get(url + obj_id, printing)
        self.assertEqual(rv.status_code, code)
        data = json.loads(rv.content)
        return data

    def del_object(self, url, obj_id, code):
        rv = self._delete(url + obj_id)
        self.assertEqual(rv.status_code, code)
        if rv.status_code != 204:
            data = json.loads(rv.content)
            return data

        else:
            return rv.status_code

    def crud_object(self, body, url):
        try:
            data = self.post_object(url, body, 202)
        except Exception as e:
            self.fail('Failure while object creation: ' + str(e))

        if url == self.url_cluster:
            crud_object = 'cluster'
            get_url = self.url_cluster_with_slash

        elif url == self.url_ngt:
            crud_object = 'node_group_template'
            get_url = self.url_ngt_with_slash

        else:
            crud_object = 'cluster_template'
            get_url = self.url_cl_tmpl_with_slash

        data = data[crud_object]
        object_id = data.get('id')

        if crud_object == 'cluster':
            cluster_state = self.get_cluster_state(object_id)
            #if cluster has no 'Active' state then cluster will be deleted
            # and raised exception with cluster state
            self.check_cluster_state(cluster_state, object_id)

        try:
            self.get_object(get_url, object_id, 200)
        except Exception as e:
            self.del_object(get_url, object_id, 204)
            self.fail('Failure while object info obtaining: ' + str(e))

        try:
            self.del_object(get_url, object_id, 204)
            if crud_object == 'cluster':
                time.sleep(5)
        except Exception as e:
            self.fail('Failure while object deletion: ' + str(e))

#----------------------make_different_template_body----------------------------

    def make_node_group_template(self, gr_name, desc, n_proc):
        processes = ['tasktracker', 'datanode']

        if n_proc == 'JT+NN':
            processes = ['jobtracker', 'namenode']

        elif n_proc == 'JT':
            processes = ['jobtracker']

        elif n_proc == 'NN':
            processes = ['namenode']

        elif n_proc == 'TT':
            processes = ['tasktracker']

        elif n_proc == 'DN':
            processes = ['datanode']

        elif n_proc == 'NN+TT+DN':
            processes = ['namenode', 'tasktracker', 'datanode']

        elif n_proc == 'JT+TT+DN':
            processes = ['jobtracker', 'tasktracker', 'datanode']

        group_template = dict(
            name=gr_name,
            description=desc,
            flavor_id=param.FLAVOR_ID,
            plugin_name=v_prm.PLUGIN_NAME,
            hadoop_version=v_prm.HADOOP_VERSION,
            node_processes=processes,
            node_configs={
                'HDFS': {},
                'MapReduce': {}
            }
        )

        return group_template

    def make_cluster_template(self, name, ngt_list):
        cluster_template = dict(
            name=name,
            plugin_name=v_prm.PLUGIN_NAME,
            hadoop_version=v_prm.HADOOP_VERSION,
            cluster_configs={},
            node_groups=[]
        )

        for key, value in ngt_list.items():
            ngt = dict(
                name='',
                node_group_template_id='',
                count=1
            )

            ngt['node_group_template_id'] = key
            ngt['count'] = value
            data = self.get_object(self.url_ngt_with_slash, key, 200)
            name = data['node_group_template']['name']
            ngt['name'] = name
            cluster_template['node_groups'].append(ngt)

        return cluster_template

    def make_cl_body_cluster_template(self, cl_tmpl_id):
        cluster_body = dict(
            name=param.CLUSTER_NAME,
            plugin_name=v_prm.PLUGIN_NAME,
            hadoop_version=v_prm.HADOOP_VERSION,
            cluster_template_id=cl_tmpl_id,
            default_image_id=v_prm.IMAGE_ID,
            user_keypair_id=param.USER_KEYPAIR_ID
        )

        return cluster_body

    def base_cl_body_node_processes(self, plugin_name, hadoop_version,
                                    user_keypair_id, image_id):
        return dict(
            name=param.CLUSTER_NAME,
            plugin_name=plugin_name,
            hadoop_version=hadoop_version,
            user_keypair_id=user_keypair_id,
            default_image_id=image_id,
            cluster_configs={},
            node_groups=[]
        )

    def add_node_group_to_cluster_body(self, cluster_body, ng_name, processes,
                                       node_count):
        cluster_body['node_groups'].append(dict(
            name=ng_name,
            flavor_id=param.FLAVOR_ID,
            node_processes=processes,
            count=node_count
        ))

    def make_hdp_cl_body_node_processes(self, node_processes):
        cluster_body = self.base_cl_body_node_processes(
            hdp_param.PLUGIN_NAME,
            hdp_param.HADOOP_VERSION,
            param.USER_KEYPAIR_ID,
            hdp_param.IMAGE_ID
        )

        for process, node_count in node_processes.items():
            if process == 'JT+NN':
                processes = ['JOBTRACKER', 'NAMENODE', 'SECONDARY_NAMENODE',
                             'GANGLIA_SERVER', 'GANGLIA_MONITOR',
                             'NAGIOS_SERVER', 'AMBARI_SERVER', 'AMBARI_AGENT']
            ng_name = 'jt-nn'

            if process == 'TT+DN':
                processes = ['TASKTRACKER', 'DATANODE', 'GANGLIA_MONITOR',
                             'HDFS_CLIENT', 'MAPREDUCE_CLIENT',
                             'AMBARI_AGENT']
                ng_name = 'tt-dn'

            self.add_node_group_to_cluster_body(cluster_body, ng_name,
                                                processes, node_count)

        return cluster_body

    def make_vanilla_cl_body_node_processes(self, node_processes):
        cluster_body = self.base_cl_body_node_processes(
            v_prm.PLUGIN_NAME,
            v_prm.HADOOP_VERSION,
            param.USER_KEYPAIR_ID,
            v_prm.IMAGE_ID
        )

        for process, node_count in node_processes.items():
            processes = ['jobtracker', 'namenode']
            ng_name = 'jt-nn'

            if process == 'TT+DN':
                processes = ['tasktracker', 'datanode']
                ng_name = 'tt-dn'

            elif process == 'JT':
                processes = ['jobtracker']
                ng_name = 'jt'

            elif process == 'NN':
                processes = ['namenode']
                ng_name = 'nn'

            elif process == 'TT':
                processes = ['tasktracker']
                ng_name = 'tt'

            elif process == 'DN':
                processes = ['datanode']
                ng_name = 'dn'

            elif process == 'JT+TT+DN':
                processes = ['jobtracker', 'tasktracker', 'datanode']
                ng_name = 'jt-tt-dn'

            elif process == 'NN+TT+DN':
                processes = ['namenode', 'tasktracker', 'datanode']
                ng_name = 'nn-tt-dn'

            self.add_node_group_to_cluster_body(cluster_body, ng_name,
                                                processes, node_count)

        return cluster_body

    def make_cl_body_node_group_templates(self, ngt_id_list):
        cluster_body = dict(
            name=param.CLUSTER_NAME,
            plugin_name=v_prm.PLUGIN_NAME,
            hadoop_version=v_prm.HADOOP_VERSION,
            user_keypair_id=param.USER_KEYPAIR_ID,
            default_image_id=v_prm.IMAGE_ID,
            cluster_configs={},
            node_groups=[]
        )

        for key, value in ngt_id_list.items():
            data = self.get_object(self.url_ngt_with_slash, key, 200)
            name = data['node_group_template']['name']
            cluster_body['node_groups'].append(dict(
                name=name,
                node_group_template_id=key,
                count=value
            ))

        return cluster_body

    def make_cl_node_processes_ngt(self, node_processes, ngt_id_list):
        cluster_body = self.make_vanilla_cl_body_node_processes(node_processes)

        for key, value in ngt_id_list.items():
            data = self.get_object(self.url_ngt_with_slash, key, 200)
            name = data['node_group_template']['name']
            cluster_body['node_groups'].append(dict(
                name=name,
                node_group_template_id=key,
                count=value
            ))

        return cluster_body

#-------------------make_node_group_template_and_get_id------------------------

    def create_node_group_templates(self):
        self.id_tt = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'worker-tt', 'qa probe', 'TT'), 202))
        self.id_jt = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'master-jt', 'qa probe', 'JT'), 202))

        self.id_nn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'master-nn', 'qa probe', 'NN'), 202))

        self.id_dn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'worker-dn', 'qa probe', 'DN'), 202))

        self.id_tt_dn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'worker-tt-dn', 'qa probe', 'TT+DN'), 202))

        self.id_jt_nn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'master-jt-nn', 'qa probe', 'JT+NN'), 202))

        self.id_nn_tt_dn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'nn-tt-dn', 'qa probe', 'NN+TT+DN'), 202))

        self.id_jt_tt_dn = self.get_object_id(
            'node_group_template', self.post_object(
                self.url_ngt, self.make_node_group_template(
                    'jt-tt-dn', 'qa probe', 'JT+TT+DN'), 202))

#---------------------delete_node_group_template-------------------------------

    def delete_node_group_templates(self):
        self.del_object(self.url_ngt_with_slash, self.id_jt_nn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_jt, 204)
        self.del_object(self.url_ngt_with_slash, self.id_nn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_tt, 204)
        self.del_object(self.url_ngt_with_slash, self.id_dn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_tt_dn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_nn_tt_dn, 204)
        self.del_object(self.url_ngt_with_slash, self.id_jt_tt_dn, 204)

#------------------------------helper_methods----------------------------------

    def check_cluster_state(self, cluster_state, cluster_id):
        if cluster_state == 'Error':
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)
            self.fail('Cluster state == \'Error\'')

        if cluster_state != 'Active':
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)
            self.fail(
                'Cluster state != \'Active\', passed %d minutes'
                % param.TIMEOUT)

    def get_cluster_state(self, object_id):
        get_data = self.get_object(self.url_cluster_with_slash, object_id, 200)
        get_data = get_data['cluster']
        i = 1
        while get_data['status'] != 'Active':
            print('GET_STATUS: ', get_data['status'])
            if get_data['status'] == 'Error':
                print('\n Data for cluster: ' + str(get_data) + '\n')

                return 'Error'

            if i > param.TIMEOUT * 6:
                print('\n Data for cluster: ' + str(get_data) + '\n')

                return get_data['status']

            get_data = self.get_object(
                self.url_cluster_with_slash, object_id, 200)
            get_data = get_data['cluster']
            time.sleep(10)
            i += 1

        return get_data['status']

    def get_object_id(self, object_type, data):
        data = data[object_type]
        return data['id']

    def ssh_connection(self, host, node_username):
        return remote.setup_ssh_connection(host, node_username,
                                           open(param.PATH_TO_SSH).read())

    def execute_command(self, host, cmd, node_username):
        with contextlib.closing(self.ssh_connection(host,
                                                    node_username)) as ssh:
            return remote.execute_command(ssh, cmd)

    def write_file_to(self, host, remote_file, data, node_username):
        with contextlib.closing(self.ssh_connection(host,
                                                    node_username)) as ssh:
            return remote.write_file_to(ssh.open_sftp(), remote_file, data)

    def read_file_from(self, host, remote_file, node_username):
        with contextlib.closing(self.ssh_connection(host,
                                                    node_username)) as ssh:
            return remote.read_file_from(ssh.open_sftp(), remote_file)

    def transfer_script_to_node(self, host, node_username,
                                script='hadoop_test/hadoop_test_script.sh'):
        self.write_file_to(str(host),
                           'script.sh',
                           open('%s/integration/%s' % (os.getcwd(),
                                                       script)).read(),
                           node_username)
        self.execute_command(str(host), 'chmod 777 script.sh', node_username)

    def try_telnet(self, host, port):
        try:
            telnetlib.Telnet(host, port)

        except Exception as e:
            self.fail('Telnet has failed: ' + str(e) +
                      '     NODE_IP: %s, PORT: %s' % (host, port))

    def create_cluster_and_get_id(self, cluster_body):
        data = self.post_object(self.url_cluster, cluster_body, 202)
        cluster_id = data['cluster']['id']

        cluster_state = self.get_cluster_state(cluster_id)
        #if cluster has no 'Active' state then cluster will be deleted
        # and raised exception with cluster state
        self.check_cluster_state(cluster_state, cluster_id)

        return cluster_id

    def create_cluster_using_ngt_and_get_id(self, cl_tmpl_id, name):
        try:
            clstr_body = self.make_cl_body_cluster_template(cl_tmpl_id)
            clstr_body['name'] = name
            return self.create_cluster_and_get_id(clstr_body)

        except Exception as e:
            self.del_object(self.url_cl_tmpl_with_slash, cl_tmpl_id, 204)
            self.fail('failure: ' + str(e))

    def get_instances_ip_and_node_processes_list(self, cluster_id):
        get_data = self.get_object(
            self.url_cluster_with_slash, cluster_id, 200, True)
        node_groups = get_data['cluster']['node_groups']

        instances_ip = {}

        for node_group in node_groups:
            instances = node_group['instances']
            for instance in instances:
                management_ip = instance['management_ip']
                instances_ip[management_ip] = \
                    node_group['node_processes']

        return instances_ip

    def get_namenode_ip_and_tt_dn_count(self, instances_ip, plugin_name):
        #This timeout is needed for deploying Hadoop services
        time.sleep(120)

        tasktracker_count = 0
        datanode_count = 0
        node_count = 0

        namenode_ip = ''

        portmap = {
            'jobtracker': param.JT_PORT,
            'namenode': param.NN_PORT,
            'tasktracker': param.TT_PORT,
            'datanode': param.DN_PORT,
            'secondary_namenode': param.SEC_NN_PORT
        }
        self.tt = 'tasktracker'
        self.dn = 'datanode'
        self.nn = 'namenode'

        if plugin_name == 'hdp':
            portmap = {
                'JOBTRACKER': param.JT_PORT,
                'NAMENODE': param.NN_PORT,
                'TASKTRACKER': param.TT_PORT,
                'DATANODE': param.DN_PORT,
                'SECONDARY_NAMENODE': param.SEC_NN_PORT
            }
            self.tt = 'TASKTRACKER'
            self.dn = 'DATANODE'
            self.nn = 'NAMENODE'

        for host, processes in instances_ip.items():
            self.try_telnet(host, '22')
            node_count += 1

            for process in processes:
                if process in portmap:
                    self.try_telnet(host, portmap[process])

            if self.tt in processes:
                tasktracker_count += 1

            if self.dn in processes:
                datanode_count += 1

            if self.nn in processes:
                namenode_ip = host

        return {
            'namenode_ip': namenode_ip,
            'tasktracker_count': tasktracker_count,
            'datanode_count': datanode_count,
            'node_count': node_count
        }

    def await_active_workers_for_namenode(self, data, node_username,
                                          hadoop_user):
        attempts_count = 100

        while True:
            active_tasktrackers_count = self.execute_command(
                data['namenode_ip'], 'sudo su -c "hadoop job \
                            -list-active-trackers" %s' % hadoop_user,
                node_username)[1]

            active_datanodes_count = int(
                self.execute_command(data['namenode_ip'],
                                     'sudo su -c "hadoop dfsadmin -report" %s \
                                     | grep "Datanodes available:.*" | awk \
                                     \'{print $3}\'' % hadoop_user,
                                     node_username)[1]
            )

            if not active_tasktrackers_count:
                active_tasktrackers_count = 0

            else:
                active_tasktrackers_count = \
                    len(active_tasktrackers_count[:-1].split('\n'))

            if (active_tasktrackers_count == data['tasktracker_count']) and (
                    active_datanodes_count == data['datanode_count']):
                break

            if attempts_count == 0:
                self.fail('Tasktracker or datanode cannot be started '
                          'within 5 minutes.')

            time.sleep(3)

            attempts_count -= 1



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base
import savanna.tests.integration.configs.parameters.common_parameters as param
import savanna.tests.integration.configs.parameters.hdp_parameters as hdp_param
import savanna.tests.integration.configs.parameters.vanilla_parameters as v_prm


class HadoopTest(base.ITestCase):

    def setUp(self):
        super(HadoopTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

    def _hadoop_testing(self, cluster_body, plugin_name, hadoop_version,
                        hadoop_user, hadoop_directory, hadoop_log_directory,
                        node_username):
        cluster_id = self.create_cluster_and_get_id(cluster_body)

        try:
            ip_instances = self.get_instances_ip_and_node_processes_list(
                cluster_id)

            clstr_info = self.get_namenode_ip_and_tt_dn_count(ip_instances,
                                                              plugin_name)
            namenode_ip = clstr_info['namenode_ip']
            node_count = clstr_info['node_count']

            self.await_active_workers_for_namenode(clstr_info,
                                                   node_username, hadoop_user)

        except Exception as e:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)
            self.fail('Failure: ' + str(e))

        try:
            for node_ip in ip_instances:
                self.transfer_script_to_node(node_ip, node_username)

        except Exception as e:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)
            self.fail('Failure while script transferring: ' + str(e))

        try:
            self.execute_command(
                namenode_ip,
                './script.sh pi -nc %s -hv %s -hd %s -hu %s -pn %s'
                % (node_count,
                   hadoop_version,
                   hadoop_directory,
                   hadoop_user,
                   plugin_name), node_username)

        except Exception as e:
            print(self.read_file_from(namenode_ip,
                                      '/tmp/outputTestMapReduce/log.txt',
                                      node_username))
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)
            self.fail('Failure while PI-job launch: ' + str(e))

        try:
            job_name = self.execute_command(namenode_ip, './script.sh gn',
                                            node_username)

            if job_name[1] == 'JobId':
                self.fail('PI-job has not been launched')

            self.assertEqual(job_name[0], 0)

        except Exception as e:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)
            self.fail('Failure while job name obtaining: ' + str(e))

        try:
            for node_ip, process_list in ip_instances.items():
                if self.tt in process_list:
                    self.assertEqual(
                        self.execute_command(
                            node_ip, './script.sh ed -jn %s -hld %s'
                            % (job_name[1][:-1], hadoop_log_directory),
                            node_username
                        )[0], 0,
                        'Host %s: not found log file of PI-job work' % node_ip)

        except Exception as e:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)
            self.fail('Failure: ' + str(e))

        try:
            self.assertEqual(
                self.execute_command(
                    namenode_ip,
                    './script.sh mr -hv %s -hd %s -hu %s -pn %s'
                    % (hadoop_version,
                       hadoop_directory,
                       hadoop_user,
                       plugin_name), node_username
                )[0], 0)

        except Exception as e:
            print(self.read_file_from(namenode_ip,
                                      '/tmp/outputTestMapReduce/log.txt',
                                      node_username))
            self.fail('Failure while HDFS check: ' + str(e))

        finally:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)

    @base.enable_test(param.ENABLE_HADOOP_TESTS_FOR_VANILLA_PLUGIN)
    def test_hadoop_single_master_for_vanilla_plugin(self):
        """This test checks Hadoop work for "Vanilla" plugin
        """
        node_processes = {'JT+NN': 1, 'TT+DN': 2}

        cluster_body = self.make_vanilla_cl_body_node_processes(node_processes)

        self._hadoop_testing(cluster_body,
                             v_prm.PLUGIN_NAME,
                             v_prm.HADOOP_VERSION,
                             v_prm.HADOOP_USER,
                             v_prm.HADOOP_DIRECTORY,
                             v_prm.HADOOP_LOG_DIRECTORY,
                             v_prm.NODE_USERNAME)

    @base.enable_test(param.ENABLE_HADOOP_TESTS_FOR_HDP_PLUGIN)
    def test_hadoop_single_master_for_hdp_plugin(self):
        """This test checks Hadoop work for "HDP" plugin
        """
        node_processes = {'JT+NN': 1, 'TT+DN': 2}

        cluster_body = self.make_hdp_cl_body_node_processes(node_processes)

        self._hadoop_testing(cluster_body,
                             hdp_param.PLUGIN_NAME,
                             hdp_param.HADOOP_VERSION,
                             hdp_param.HADOOP_USER,
                             hdp_param.HADOOP_DIRECTORY,
                             hdp_param.HADOOP_LOG_DIRECTORY,
                             hdp_param.NODE_USERNAME)



# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import savanna.exceptions as e
from savanna.plugins.general import exceptions as ex
from savanna.plugins.general import utils


class Validator(object):
    def validate(self, cluster):
        funcs = inspect.getmembers(Validator, predicate=inspect.ismethod)
        for func in funcs:
            if func[0].startswith("check_"):
                getattr(self, func[0])(cluster)

    def _get_named_node_group(self, cluster, ng_name):
        return next((ng for ng in cluster.node_groups
                     if ng.name == ng_name), None)

    def validate_scaling(self, cluster, existing, additional):
        orig_existing_count = {}
        try:
            for ng_name in existing:
                node_group = self._get_named_node_group(cluster, ng_name)
                if node_group:
                    orig_existing_count[ng_name] = node_group.count
                    node_group.count = int(existing[ng_name])
                else:
                    raise RuntimeError('Node group not found: {0}'.format(
                        ng_name
                    ))
            for node_group in additional:
                node_group.count = additional[node_group]
                cluster.node_groups.append(node_group)

            self.validate(cluster)

        finally:
            for node_group in additional:
                node_group.count = 0
                cluster.node_groups.remove(node_group)
            for ng_name in orig_existing_count:
                node_group = self._get_named_node_group(cluster, ng_name)
                node_group.count = orig_existing_count[ng_name]

    def check_for_namenode(self, cluster):
        count = sum([ng.count for ng
                     in utils.get_node_groups(cluster, "NAMENODE")])
        if count != 1:
            raise ex.NotSingleNameNodeException(count)

    def check_for_jobtracker_and_tasktracker(self, cluster):
        jt_count = sum([ng.count for ng
                        in utils.get_node_groups(cluster, "JOBTRACKER")])

        if jt_count not in [0, 1]:
            raise ex.NotSingleJobTrackerException(jt_count)

        tt_count = sum([ng.count for ng
                        in utils.get_node_groups(cluster, "TASKTRACKER")])
        if jt_count is 0 and tt_count > 0:
            raise ex.TaskTrackersWithoutJobTracker()

    def check_for_ambari_server(self, cluster):
        count = sum([ng.count for ng
                     in utils.get_node_groups(cluster, "AMBARI_SERVER")])
        if count != 1:
            raise NotSingleAmbariServerException(count)

    def check_for_ambari_agents(self, cluster):
        for ng in cluster.node_groups:
            if "AMBARI_AGENT" not in ng.node_processes:
                raise AmbariAgentNumberException(ng.name)


class NoNameNodeException(e.SavannaException):
    def __init__(self):
        message = ("Hadoop cluster should contain at least one namenode")
        code = "NO_NAMENODE"

        super(NoNameNodeException, self).__init__(message, code)


class NotSingleAmbariServerException(e.SavannaException):
    def __init__(self, count):
        message = ("Hadoop cluster should contain 1 Ambari Server "
                   "instance. Actual Ambari server count is %s" % count)
        code = "NOT_SINGLE_AMBARI_SERVER"

        super(NotSingleAmbariServerException, self).__init__(message, code)


class AmbariAgentNumberException(e.SavannaException):
    def __init__(self, count):
        message = ("Hadoop cluster should have an ambari agent per "
                   "node group. Node group %s has no Ambari Agent" % count)
        code = "WRONG_NUMBER_AMBARI_AGENTS"

        super(AmbariAgentNumberException, self).__init__(message, code)



# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import requests
from savanna import context
from savanna import exceptions as exc
from savanna.openstack.common import jsonutils as json
from savanna.openstack.common import log as logging
from savanna.plugins.hdp import blueprintprocessor as bp
from savanna.plugins.hdp import clusterspec
from savanna.plugins.hdp import configprovider as cfg
from savanna.plugins.hdp import exceptions as ex
from savanna.plugins.hdp import hadoopserver as h
from savanna.plugins.hdp import savannautils as s
from savanna.plugins.hdp import validator as v
from savanna.plugins import provisioning as p

LOG = logging.getLogger(__name__)


class AmbariPlugin(p.ProvisioningPluginBase):
    def __init__(self):
        self.cluster_ambari_mapping = {}
        self.default_config = self._get_default_cluster_configuration()

    def create_cluster(self, cluster, cluster_template):

        if cluster_template is None:
            raise ValueError('must supply cluster template')

        cluster_spec = clusterspec.ClusterSpec(
            cluster_template, cluster=cluster)

        hosts = self._get_servers(cluster)
        ambari_info = self.get_ambari_info(cluster_spec)
        self.cluster_ambari_mapping[cluster.name] = ambari_info
        ambari_uri = self._get_ambari_uri(cluster_spec)

        servers = []
        for host in hosts:
            servers.append(
                h.HadoopServer(host, cluster_spec.node_groups[host.role],
                               ambari_uri=ambari_uri))

        provisioned = self._provision_cluster(
            cluster.name, cluster_spec, ambari_info, servers)

        if provisioned:
            installed = self._install_services(cluster.name, ambari_info)
            if installed:
                LOG.info("Install of Hadoop stack successful.")
                # add service urls
                self._set_cluster_info(cluster, cluster_spec, ambari_info)
            else:
                raise ex.HadoopProvisionError(
                    'Installation of Hadoop stack failed.')

        else:
            raise ex.HadoopProvisionError(
                'Provisioning of Hadoop cluster failed.')

    def _get_servers(self, cluster):
        servers = []
        if hasattr(cluster, 'node_groups') and cluster.node_groups is not None:
            # code for a savanna cluster object
            for node_group in cluster.node_groups:
                for server in node_group.instances:
                    setattr(server, 'role', node_group.name)
                    setattr(server, 'node_processes',
                            node_group.node_processes)
                    servers.append(server)
        else:
            # cluster is actually a cloud context
            servers = cluster.instances

        return servers

    def get_node_processes(self, hadoop_version):
        #TODO(jmaron): use version information
        node_processes = {}
        for service in self.default_config.services:
            components = []
            for component in service.components:
                components.append(component.name)
            node_processes[service.name] = components

        return node_processes

    def convert(self, cluster_template, config):
        normalized_config = clusterspec.ClusterSpec(config).normalize()

        #TODO(jspeidel):  can we get the name (first arg) from somewhere?

        cluster_template = s.convert(cluster_template, normalized_config,
                                     self.get_configs(
                                         cluster_template.hadoop_version))

        return cluster_template

    def update_infra(self, cluster):
        for node_group in cluster.node_groups:
            node_group.image = cluster.default_image_id

    def convert_props_to_template(self, props):
        raise NotImplementedError('not yet supported')

    def _add_cluster(self, ambari_info, name):
        add_cluster_url = 'http://{0}/api/v1/clusters/{1}'.format(
            ambari_info.get_address(), name)
        #TODO(jspeidel): get stack info from config spec
        result = requests.post(add_cluster_url,
                               data='{"Clusters": {"version" : "HDP-1.3.0"}}',
                               auth=(ambari_info.user, ambari_info.password))

        if result.status_code != 201:
            LOG.warning(
                'Create cluster command failed. {0}'.format(result.text))
            return False

        return True

    def _add_configurations_to_cluster(
            self, cluster_spec, ambari_info, name):

        configs = cluster_spec.configurations
        config_url = 'http://{0}/api/v1/clusters/{1}'.format(
            ambari_info.get_address(), name)

        body = {}
        clusters = {}
        body['Clusters'] = clusters
        for config_name in configs:
            if config_name == 'ambari':
                continue

            config_body = {}
            clusters['desired_config'] = config_body
            config_body['type'] = config_name
            #TODO(jspeidel): hard coding for now
            config_body['tag'] = 'v1'
            config_body['properties'] = configs[config_name]
            result = requests.put(config_url, data=json.dumps(body), auth=(
                ambari_info.user, ambari_info.password))
            if result.status_code != 200:
                LOG.warning(
                    'Set configuration command failed. {0}'.format(
                        result.text))
                return False

        return True

    def _add_services_to_cluster(self, cluster_spec, ambari_info, name):
        services = cluster_spec.services
        add_service_url = 'http://{0}/api/v1/clusters/{1}/services/{2}'
        for service in services:
            if service.name != 'AMBARI':
                result = requests.post(add_service_url.format(
                    ambari_info.get_address(), name, service.name),
                    auth=(ambari_info.user, ambari_info.password))
                if result.status_code != 201:
                    LOG.warning(
                        'Create service command failed. {0}'.format(
                            result.text))
                    return False

        return True

    def _add_components_to_services(self, cluster_spec, ambari_info, name):
        add_component_url = 'http://{0}/api/v1/clusters/{1}/services/{' \
                            '2}/components/{3}'
        for service in cluster_spec.services:
            if service.name != 'AMBARI':
                for component in service.components:
                    result = requests.post(add_component_url.format(
                        ambari_info.get_address(), name, service.name,
                        component.name),
                        auth=(ambari_info.user, ambari_info.password))
                    if result.status_code != 201:
                        LOG.warning(
                            'Create component command failed. {0}'.format(
                                result.text))
                        return False

        return True

    def _add_hosts_and_components(
            self, cluster_spec, servers, ambari_info, name):

        add_host_url = 'http://{0}/api/v1/clusters/{1}/hosts/{2}'
        add_host_component_url = 'http://{0}/api/v1/clusters/{1}' \
                                 '/hosts/{2}/host_components/{3}'
        for host in servers:
            hostname = host.instance.fqdn.lower()
            result = requests.post(
                add_host_url.format(ambari_info.get_address(), name, hostname),
                auth=(ambari_info.user, ambari_info.password))
            if result.status_code != 201:
                LOG.warning(
                    'Create host command failed. {0}'.format(result.text))
                return False

            node_group_name = host.node_group.name
            #TODO(jspeidel): ensure that node group exists
            node_group = cluster_spec.node_groups[node_group_name]
            for component in node_group.components:
                # don't add any AMBARI components
                if component.find('AMBARI') != 0:
                    result = requests.post(add_host_component_url.format(
                        ambari_info.get_address(), name, hostname, component),
                        auth=(ambari_info.user, ambari_info.password))
                    if result.status_code != 201:
                        LOG.warning(
                            'Create host_component command failed. {0}'.format(
                                result.text))
                        return False

        return True

    def _provision_cluster(self, name, cluster_spec, ambari_info, servers):
        #TODO(jspeidel): encapsulate in another class

        LOG.info('Provisioning Cluster via Ambari Server: {0} ...'.format(
            ambari_info.get_address()))

        for server in servers:
            context.spawn("hdp-provision-instance-%s" %
                          server.instance.hostname,
                          server.provision_ambari, ambari_info)

        self._wait_for_host_registrations(len(servers), ambari_info)
        self._set_ambari_credentials(cluster_spec, ambari_info)

        # add cluster
        if not self._add_cluster(ambari_info, name):
            return False

        # add configurations to cluster
        if not self._add_configurations_to_cluster(cluster_spec,
                                                   ambari_info, name):
            return False

        # add services
        if not self._add_services_to_cluster(cluster_spec,
                                             ambari_info, name):
            return False

        # add components to services
        if not self._add_components_to_services(cluster_spec,
                                                ambari_info, name):
            return False

        # add hosts and host_components
        if not self._add_hosts_and_components(cluster_spec, servers,
                                              ambari_info, name):
            return False

        return True

    def _wait_for_host_registrations(self, num_hosts, ambari_info):
        LOG.info(
            'Waiting for all Ambari agents to register with server ...')

        url = 'http://{0}/api/v1/hosts'.format(ambari_info.get_address())
        result = None
        json_result = None

        #TODO(jspeidel): timeout
        while result is None or len(json_result['items']) < num_hosts:
            context.sleep(5)
            try:
                result = requests.get(url, auth=(ambari_info.user,
                                                 ambari_info.password))
                json_result = json.loads(result.text)

                # TODO(jspeidel): just for debug
                LOG.info('Registered Hosts: {0} of {1}'.format(
                    len(json_result['items']), num_hosts))
                for hosts in json_result['items']:
                    LOG.debug('Registered Host: {0}'.format(
                        hosts['Hosts']['host_name']))
            except requests.ConnectionError:
                #TODO(jspeidel): max wait time
                LOG.info('Waiting to connect to ambari server ...')

    def _install_services(self, cluster_name, ambari_info):
        LOG.info('Installing required Hadoop services ...')

        ambari_address = ambari_info.get_address()
        install_url = 'http://{0}/api/v1/clusters/{' \
                      '1}/services?ServiceInfo/state=INIT'.format(
                          ambari_address, cluster_name)
        body = '{"ServiceInfo": {"state" : "INSTALLED"}}'

        result = requests.put(install_url, data=body, auth=(
            ambari_info.user, ambari_info.password))

        if result.status_code == 202:
            #TODO(jspeidel) don't hard code request id
            success = self._wait_for_async_request(
                self._get_async_request_uri(ambari_info, cluster_name, 1),
                auth=(ambari_info.user, ambari_info.password))
            if success:
                LOG.info("Install of Hadoop stack successful.")
                self._finalize_ambari_state(ambari_info)
            else:
                LOG.critical('Install command failed.')
                raise RuntimeError('Hadoop service install failed')
        else:
            LOG.error(
                'Install command failed. {0}'.format(result.text))
            raise RuntimeError('Hadoop service install failed')

        return success

    def _get_async_request_uri(self, ambari_info, cluster_name, request_id):
        return 'http://{0}/api/v1/clusters/{1}/requests/{' \
               '2}/tasks?fields=Tasks/status'.format(
               ambari_info.get_address(), cluster_name,
               request_id)

    def _wait_for_async_request(self, request_url, auth):
        started = False
        while not started:
            result = requests.get(request_url, auth=auth)
            LOG.debug(
                'async request ' + request_url + ' response:\n' + result.text)
            json_result = json.loads(result.text)
            started = True
            for items in json_result['items']:
                status = items['Tasks']['status']
                if status == 'FAILED' or status == 'ABORTED':
                    return False
                else:
                    if status != 'COMPLETED':
                        started = False

            context.sleep(5)
        return started

    def _start_services(self, cluster_name, ambari_info):
        LOG.info('Starting Hadoop services ...')
        LOG.info('Cluster name: {0}, Ambari server address: {1}'
                 .format(cluster_name, ambari_info.get_address()))
        start_url = 'http://{0}/api/v1/clusters/{1}/services?ServiceInfo/' \
                    'state=INSTALLED'.format(
                        ambari_info.get_address(), cluster_name)
        body = '{"ServiceInfo": {"state" : "STARTED"}}'

        auth = (ambari_info.user, ambari_info.password)
        result = requests.put(start_url, data=body, auth=auth)
        if result.status_code == 202:
            # don't hard code request id
            success = self._wait_for_async_request(
                self._get_async_request_uri(ambari_info, cluster_name, 2),
                auth=auth)
            if success:
                LOG.info(
                    "Successfully started Hadoop cluster '{0}'.".format(
                        cluster_name))
            else:
                LOG.critical('Failed to start Hadoop cluster.')
                raise RuntimeError('Failed to start Hadoop cluster.')

        else:
            LOG.critical(
                'Start command failed. Status: {0}, response: {1}'.
                format(result.status_code, result.text))
            raise RuntimeError('Hadoop cluster start failed.')

    def _install_components(self, ambari_info, auth, cluster_name, servers):
        LOG.info('Starting Hadoop components while scaling up')
        LOG.info('Cluster name {0}, Ambari server ip {1}'
                 .format(cluster_name, ambari_info.get_address()))
        # query for the host components on the given hosts that are in the
        # INIT state
        body = '{"HostRoles": {"state" : "INSTALLED"}}'
        install_uri = 'http://{0}/api/v1/clusters/{' \
                      '1}/host_components?HostRoles/state=INIT&' \
                      'HostRoles/host_name.in({2})'.format(
                      ambari_info.get_address(),
                      cluster_name,
                      self._get_host_list(servers))
        self._exec_ambari_command(auth, body, install_uri)

    def _start_components(self, ambari_info, auth, cluster_name, servers):
        # query for all the host components on one of the hosts in the
        # INSTALLED state, then get a list of the client services in the list
        installed_uri = 'http://{0}/api/v1/clusters/{' \
                        '1}/host_components?HostRoles/state=INSTALLED&' \
                        'HostRoles/host_name.in({2})' \
            .format(ambari_info.get_address(), cluster_name,
                    self._get_host_list(servers))
        result = requests.get(installed_uri, auth=auth)
        if result.status_code == 200:
            LOG.debug(
                'GET response: {0}'.format(result.text))
            json_result = json.loads(result.text)
            items = json_result['items']
            # select non-CLIENT items
            inclusion_list = list(set([x['HostRoles']['component_name']
                                       for x in items if "CLIENT" not in
                                       x['HostRoles']['component_name']]))

            # query and start all non-client components on the given set of
            # hosts
            body = '{"HostRoles": {"state" : "STARTED"}}'
            start_uri = 'http://{0}/api/v1/clusters/{' \
                        '1}/host_components?HostRoles/state=INSTALLED&' \
                        'HostRoles/host_name.in({2})' \
                        '&HostRoles/component_name.in({3})'.format(
                        ambari_info.get_address(), cluster_name,
                        self._get_host_list(servers),
                        ",".join(inclusion_list))
            self._exec_ambari_command(auth, body, start_uri)
        else:
            raise RuntimeError('Unable to determine installed service '
                               'components in scaled instances.  status'
                               ' code returned = {0}'.format(result.status))

    def _install_and_start_components(self, cluster_name, servers,
                                      ambari_info):
        auth = (ambari_info.user, ambari_info.password)

        self._install_components(ambari_info, auth, cluster_name, servers)

        self._start_components(ambari_info, auth, cluster_name, servers)

    def _exec_ambari_command(self, auth, body, cmd_uri):

        LOG.debug('PUT URI: {0}'.format(cmd_uri))
        result = requests.put(cmd_uri, data=body,
                              auth=auth)
        if result.status_code == 202:
        # don't hard code request id
            LOG.debug(
                'PUT response: {0}'.format(result.text))
            json_result = json.loads(result.text)
            href = json_result['href'] + '/tasks?fields=Tasks/status'
            success = self._wait_for_async_request(href, auth)
            if success:
                LOG.info(
                    "Successfully changed state of Hadoop components ")
            else:
                LOG.critical('Failed to change state of Hadoop '
                             'components')
                raise RuntimeError('Failed to change state of Hadoop '
                                   'components')

        else:
            LOG.error(
                'Command failed. Status: {0}, response: {1}'.
                format(result.status_code, result.text))
            raise RuntimeError('Hadoop/Ambari command failed.')

    def _get_default_cluster_configuration(self):
        with open(os.path.join(os.path.dirname(__file__), 'resources',
                               'default-cluster.template'), 'r') as f:
            return clusterspec.ClusterSpec(f.read())

    def _set_cluster_info(self, cluster, cluster_spec, ambari_info):
        info = cluster.info

        try:
            jobtracker_ip = cluster_spec.determine_host_for_server_component(
                'JOBTRACKER').management_ip
        except Exception:
            pass
        else:
            info['MapReduce'] = {
                'Web UI': 'http://%s:50030' % jobtracker_ip
            }

        try:
            namenode_ip = cluster_spec.determine_host_for_server_component(
                'NAMENODE').management_ip
        except Exception:
            pass
        else:
            info['HDFS'] = {
                'Web UI': 'http://%s:50070' % namenode_ip
            }

        info['Ambari Console'] = {
            'Web UI': 'http://%s' % ambari_info.get_address()
        }

    def _finalize_ambari_state(self, ambari_info):
        LOG.info('Finalizing Ambari cluster state.')

        persist_state_uri = 'http://{0}/api/v1/persist'.format(
            ambari_info.get_address())
        # this post data has non-standard format because persist
        # resource doesn't comply with Ambari API standards
        persist_data = '{ "CLUSTER_CURRENT_STATUS":' \
                       '"{\\"clusterState\\":\\"CLUSTER_STARTED_5\\"}" }'
        result = requests.post(persist_state_uri, data=persist_data,
                               auth=(ambari_info.user, ambari_info.password))

        if result.status_code != 201 and result.status_code != 202:
            LOG.warning('Finalizing of Ambari cluster state failed. {0}'.
                        format(result.text))
            raise ex.HadoopProvisionError('Unable to finalize Ambari state.')

    def _set_ambari_credentials(self, cluster_spec, ambari_info):
        services = cluster_spec.services
        for service in services:
            if service.name == 'AMBARI':
                is_admin_provided = False
                admin_user = ambari_info.user
                admin_password = ambari_info.password
                for u in service.users:
                    if u.name == 'admin':
                        self._update_ambari_admin_user(
                            u.password, ambari_info)
                        is_admin_provided = True
                        ambari_info.user = 'admin'
                        ambari_info.password = u.password
                    else:
                        self._add_ambari_user(u, ambari_info)
                        if 'admin' in u.groups:
                            admin_user = u.name
                            admin_password = u.password

                if not is_admin_provided:
                    if admin_user is None:
                        raise ex.HadoopProvisionError("An Ambari user in the "
                                                      "admin group must be "
                                                      "configured.")
                    ambari_info.user = admin_user
                    ambari_info.password = admin_password
                    self._delete_ambari_user('admin', ambari_info)
                break

    def _update_ambari_info_credentials(self, cluster_spec, ambari_info):
        services = cluster_spec.services
        ambari_service = next((service for service in services if
                               service.name == 'AMBARI'), None)
        if ambari_service is not None:
            admin_user = next((user for user in ambari_service.users
                               if 'admin' in user.groups), None)
            if admin_user is not None:
                ambari_info.user = admin_user.name
                ambari_info.password = admin_user.password

        LOG.info('Using "{0}" as admin user for scaling of cluster'
                 .format(ambari_info.user))

    def _update_ambari_admin_user(self, password, ambari_info):
        old_pwd = ambari_info.password
        user_url = 'http://{0}/api/v1/users/admin'.format(
            ambari_info.get_address())
        update_body = '{{"Users":{{"roles":"admin,user","password":"{0}",'\
                      '"old_password":"{1}"}} }}'.format(password, old_pwd)

        request = self._get_rest_request()
        result = request.put(user_url, data=update_body, auth=(
            ambari_info.user, ambari_info.password))

        if result.status_code != 200:
            raise ex.HadoopProvisionError('Unable to update Ambari admin user'
                                          ' credentials: {0}'.
                                          format(result.text))

    def _add_ambari_user(self, user, ambari_info):
        user_url = 'http://{0}/api/v1/users/{1}'.format(
            ambari_info.get_address(), user.name)

        create_body = '{{"Users":{{"password":"{0}","roles":"{1}"}} }}'.\
            format(user.password, '%s' % ','.join(map(str, user.groups)))

        request = self._get_rest_request()
        result = request.post(user_url, data=create_body, auth=(
            ambari_info.user, ambari_info.password))

        if result.status_code != 201:
            raise ex.HadoopProvisionError('Unable to create Ambari user: {0}'.
                                          format(result.text))

    def _delete_ambari_user(self, user_name, ambari_info):
        user_url = 'http://{0}/api/v1/users/{1}'.format(
            ambari_info.get_address(), user_name)

        request = self._get_rest_request()
        result = request.delete(user_url, auth=(
            ambari_info.user, ambari_info.password))

        if result.status_code != 200:
            raise ex.HadoopProvisionError('Unable to delete Ambari user: {0}'
                                          ' : {1}'.format(user_name,
                                          result.text))

    def _get_rest_request(self):
        return requests

    # SAVANNA PLUGIN SPI METHODS:
    def _get_blueprint_processor(self, cluster):
        processor = bp.BlueprintProcessor(json.load(
            open(os.path.join(os.path.dirname(__file__), 'resources',
                              'default-cluster.template'), "r")))
        processor.process_user_inputs(self._map_to_user_inputs(
            '1.3.0', cluster.cluster_configs))
        processor.process_node_groups(cluster.node_groups)
        return processor

    def configure_cluster(self, cluster):
        # take the user inputs from the cluster and node groups and convert
        # to a ambari blueprint
        processor = self._get_blueprint_processor(cluster)
        # NOTE: for the time being we are going to ignore the node group
        # level configurations.  we are not currently
        # defining node level configuration items (i.e. scope='cluster' in
        # all cases for returned configs)

        #create a cloud context

        #TODO(jmaron):  is base host name really necessary any longer?
        #cloud_ctx = ClusterContext(None, LOG)
        #self._add_instances_to_cluster_context (cloud_ctx, cluster)

        self.create_cluster(cluster, json.dumps(processor.blueprint))

    def get_versions(self):
        return ['1.3.0']

    def get_configs(self, hadoop_version):
        config_resource = cfg.ConfigurationProvider(
            json.load(open(os.path.join(os.path.dirname(__file__), 'resources',
                                        'ambari-config-resource.json'), "r")))
        return config_resource.get_config_items()

    # cluster name argument supports the non-savanna cluster creation mode
    def start_cluster(self, cluster, cluster_name=None):
        if cluster_name is None:
            cluster_name = cluster.name

        self._start_services(
            cluster_name, self.cluster_ambari_mapping[cluster_name])

    def get_title(self):
        return 'Hortonworks Data Platform'

    def setup(self, conf):
        super(AmbariPlugin, self).setup(conf)

    def get_description(self):
        return 'The Hortonworks OpenStack plugin works with project ' \
               'Savanna to automate the deployment of the Hortonworks data' \
               ' platform on OpenStack based public & private clouds'

    def validate(self, cluster):
        validator = v.Validator()
        validator.validate(cluster)

    def scale_cluster(self, cluster, instances):
        processor = self._get_blueprint_processor(cluster)
        cluster_spec = clusterspec.ClusterSpec(
            json.dumps(processor.blueprint), cluster=cluster)
        ambari_uri = self._get_ambari_uri(cluster_spec)

        servers = []
        for instance in instances:
            host_role = s.get_host_role(instance)
            servers.append(h.HadoopServer(instance,
                                          cluster_spec.node_groups
                                          [host_role],
                                          ambari_uri=ambari_uri))

        ambari_info = self.get_ambari_info(cluster_spec)
        self._update_ambari_info_credentials(cluster_spec, ambari_info)

        for server in servers:
            context.spawn("hdp-scaling-instance-%s" %
                          server.instance.hostname,
                          server.provision_ambari, ambari_info)

        self._wait_for_host_registrations(self._get_num_hosts(cluster),
                                          ambari_info)

        #  now add the hosts and the component
        self._add_hosts_and_components(cluster_spec, servers,
                                       ambari_info, cluster.name)

        self._install_and_start_components(cluster.name, servers, ambari_info)

    def decommission_nodes(self, cluster, instances):
        raise exc.InvalidException('The HDP plugin does not yet support the '
                                   'decommissioning of nodes')

    def validate_scaling(self, cluster, existing, additional):
        # see if additional servers are slated for "MASTER" group
        validator = v.Validator()
        validator.validate_scaling(cluster, existing, additional)

    def _get_num_hosts(self, cluster):
        count = 0
        for node_group in cluster.node_groups:
            count += node_group.count

        return count

    def _get_host_list(self, servers):
        host_list = [server.instance.fqdn.lower() for server in servers]
        return ",".join(host_list)

    def _get_ambari_uri(self, cluster_spec):
        ambari_config = cluster_spec.configurations['ambari']
        return ambari_config.get('repo.uri', None)

    def get_ambari_info(self, cluster_spec):
        ambari_host = cluster_spec.determine_host_for_server_component(
            'AMBARI_SERVER')

        port = cluster_spec.configurations['ambari'].get(
            'server.port', '8080')

        return AmbariInfo(ambari_host, port, 'admin', 'admin')


class AmbariInfo():
    def __init__(self, host, port, user, password):
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    def get_address(self):
        return '{0}:{1}'.format(self.host.management_ip, self.port)
